#ifndef HDR_GAMESTATE_HPP__
#define HDR_GAMESTATE_HPP__

#include "GameState.hpp"

#include <cstdint>
#include <deque>
#include <functional>
#include <vector>
#include <set>
#include <string>
#include <utility>

namespace game_state
{
	enum TurnResult
	{
		Continue, Next, Lose, Win
	};

	enum class ShifterShape
	{
		Cat, Wolf, Turtle, Bird,         // << Accessible to Player
		Rat, Minotaur, Cultist, Dragon,  // << Accessible to Opponents
		Forrest, Wall, Dungeon,          // << Accesible to Environments
		Wizard                           // << Accessible to Wizard
	};

	enum class Modifier
	{
		Dry, Wet, Cold, Hot, Physical, Magical
	};

	enum class StackAction
	{
		Attack, // can be physical or magical
		Defend,
		Dry,
		Wet,
		Cold,
		Hot,
		//Harm,  // --> is really Attack, Magical
		Heal,
		EnvNormal,
		EnvRandom,
		ResistCold,
		ResistHot,
		ResistWet,
		WeakToCold,
		WeakToHot,
		WeakToWet,
		LessOffence,
		MoreOffence,
		LessDefence,
		MoreDefence,
		Idle,
	};

	//enum class StackActionTarget
	//{
	//	Self, Other, Environment
	//}; // ---> Should maybe be: Player, Opponent, Environment?

	enum class TurnChoice
	{
		NormalAttack, Evade, ShiftToCat, ShiftToWolf, ShiftToTurtle, ShiftToBird, Special, Suggest, OoC
	};

	typedef std::function<StackAction(const uint32_t&)> behave_func_t;

	void CleanGlobals();
	std::pair<TurnResult, std::string>  HandlePlayerChoice(const TurnChoice& player_choice);
	std::pair<TurnResult, std::string>  HandleWizardChoice();
	std::pair<TurnResult, std::string>  HandleOppoChoice();
	std::pair<TurnResult, std::string>  HandleTurnChange();

	class ActionStack
	{
	public:
		ActionStack(const behave_func_t& func_get_next);

		void Fill(const uint32_t& num);
		StackAction Peek(const uint32_t& num) const;
		StackAction Pop(const uint32_t& turn);
		void Confuse();
		void ReplaceLast(const StackAction& action);

	private:
		int32_t fill_;
		std::deque<StackAction> actions_;
		behave_func_t func_get_next_;
	};

	class BaseCreature
	{
	public:
		BaseCreature();
		BaseCreature(const behave_func_t& func_get_next);

		virtual int32_t GetCurrentAttack() const;
		virtual int32_t GetCurrentDefence() const;
		virtual int32_t GetCurrentHitPoints() const;
		void ResetStats();
		float GetPercentage();

		virtual bool NextTurn(); // NOTE: returns dead or not!
		virtual std::string HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength = 1); // NOTE: returns message for player
		virtual std::string PlayerEffect(const TurnChoice& player_choice);
		virtual ActionStack* GetActionStack();
		virtual Modifier GetOffensiveModifier();

		virtual ShifterShape GetShape() = 0; // NOTE: Pure virtual!
		virtual bool TryShapeshift(const ShifterShape& shape) = 0; // NOTE: Pure virtual!
		
	protected:
		void SetBaseStats(const int32_t& hp, const int32_t& atk, const int32_t& def);

		virtual float ApplyModifier(const Modifier& env); // NOTE: > 1.f strength, < 1.f weakness

		int32_t hp_max_;
		int32_t hp_left_;
		int32_t atk_norm_;
		int32_t atk_turn_;
		int32_t def_norm_;
		int32_t def_turn_;

		std::set<Modifier> weaknesses_norm_;
		std::set<Modifier> weaknesses_turn_;
		std::set<Modifier> strengths_norm_;
		std::set<Modifier> strengths_turn_;

		ActionStack next_actions_;

		/** FUTURE:
		int32_t speed_;
		**/

	private:
		static StackAction StayIdle(const uint32_t&);
	};

	class Player : public BaseCreature
	{
	public:
		static Player* GetInstance();

		virtual ShifterShape GetShape();
		virtual bool TryShapeshift(const ShifterShape& shape);
		virtual std::string PlayerEffect(const TurnChoice& player_choice);

		std::string Describe(const TurnChoice& player_choice) const;

	private:
		Player();
		ShifterShape shape_;
		//uint32_t skip_turns_; // no, too complicated
	};

	class Wizard : public BaseCreature
	{
	public:
		static Wizard* GetInstance();

		virtual std::string HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength = 1);
		virtual bool NextTurn(); // NOTE: returns dead or not! : wiz doenst die in current design
		virtual ShifterShape GetShape();
		virtual bool TryShapeshift(const ShifterShape& shape);
		virtual std::string PlayerEffect(const TurnChoice& player_choice);
		virtual Modifier GetOffensiveModifier();

	private:
		Wizard();

		static StackAction Behaviour(const uint32_t& turn);
	};

	class EnvironmentG : public BaseCreature
	{
	public:
		static EnvironmentG* GetInstance();

		virtual bool NextTurn(); // NOTE: returns dead or not! : env cant die, always true
		virtual std::string HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength = 1); // NOTE: returns message for player : env mods only!
		virtual ShifterShape GetShape();
		virtual bool TryShapeshift(const ShifterShape& shape);
		//virtual std::string PlayerEffect(const TurnChoice& player_choice); // Not nesc??

		const std::set<Modifier> GetModifiers() const;

	private:
		EnvironmentG();

		Modifier humidity_;
		Modifier temperature_;
		ShifterShape looks_;
	};

	class CurrentOpponent : public BaseCreature
	{
	public:
		static CurrentOpponent* GetInstance();

		virtual int32_t CurrentOpponent::GetCurrentAttack() const;
		virtual int32_t CurrentOpponent::GetCurrentDefence() const;
		virtual int32_t CurrentOpponent::GetCurrentHitPoints() const;
		
		virtual std::string HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength = 1); // NOTE: returns message for player : env mods only!
		virtual bool NextTurn(); // NOTE: returns dead or not! : yes, 
		virtual ActionStack* GetActionStack();
		virtual ShifterShape GetShape();
		virtual bool TryShapeshift(const ShifterShape& shape);
		virtual std::string PlayerEffect(const TurnChoice& player_choice);
		virtual Modifier GetOffensiveModifier();

		float GetPercentageX();

		bool FinalDeath();
		
	private:
		CurrentOpponent();

		BaseCreature* current_;
		std::deque<BaseCreature*> sorting_algo_of_evil_;
	};

	class SingleOppo : public BaseCreature
	{
	public:
		SingleOppo(const ShifterShape& looks, const behave_func_t& behave_func);

		virtual ShifterShape GetShape();
		virtual bool TryShapeshift(const ShifterShape& shape);
		virtual std::string PlayerEffect(const TurnChoice& player_choice);
		virtual Modifier GetOffensiveModifier();

		static StackAction RatBehaviourA(const uint32_t& turn);
		static StackAction MinotaurBehaviourA(const uint32_t& turn);
		static StackAction CultistBehaviourA(const uint32_t& turn);
		static StackAction DragonBehaviourA(const uint32_t& turn);

	private:
		ShifterShape looks_;
		//behave_func_t behave_func_;
	};
}

#endif // HDR_GAMESTATE_HPP__
